{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE TypeSynonymInstances #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE BangPatterns #-}

-- | These terminals provide individual bit unsetting capability and
-- permutations of set bits.
--
-- This allows to write
-- @f <<< UnsetBit % SetPermutations % table@
-- to highlight an individual bit, then try all permutations of the
-- remaining bits, and look up the optimal score of said bitset. Can be
-- used to implement TSP.

module ADP.Fusion.Permutation where

import           Data.Strict.Tuple
import           Data.Array.Repa.Index
import qualified Data.Vector.Fusion.Stream.Monadic as M
import           Data.Vector.Fusion.Stream.Size
import           Data.Bits
import           Data.List (permutations)

import ADP.Fusion.Classes

import Data.Array.Repa.Index.BitSet



-- | Given a 'BitSet', this terminal symbol will provide each individual
-- @1@ bit in turn -- or rather their position in the bitset. I.e if the
-- value @k@ is returned, then the @k@'th bit is set. The bit is unset
-- further to the right of this terminal!
--
-- TODO the way the individual bits to be unset are generated is probably
-- horribly slow

data UnsetBit = UnsetBit

instance Build UnsetBit

instance (Element ls TspIndex) => Element (ls :!: UnsetBit) TspIndex where
    type Arg (ls :!: UnsetBit) = Arg ls :. Int

instance (Monad m, Element ls TspIndex, MkStream m ls TspIndex) => MkStream m (ls :!: UnsetBit) TspIndex where

data Final = Final

instance (Element ls TspIndex) => Element (ls :!: Final) TspIndex where
    type Arg (ls :!: Final) = Arg ls :. Int

instance (Monad m, Element ls TspIndex, MkStream m ls TspIndex) => MkStream m (ls :!: Final) TspIndex where


{-
instance (Element ls BitSet) => Element (ls :!: UnsetBit) BitSet where
    data Elm (ls :!: UnsetBit) BitSet = ElmUnsetBit !Int !BitSet !(Elm ls BitSet)
    type Arg (ls :!: UnsetBit)        = Arg ls :. Int
    getArg (ElmUnsetBit x _ ls)       = getArg ls :. x
    getIdx (ElmUnsetBit _ i _ )       = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance (Monad m, Element ls BitSet, MkStream m ls BitSet) => MkStream m (ls :!: UnsetBit) BitSet where
    mkStream (ls :!: UnsetBit) sv ix = M.flatten mk step Unknown $ mkStream ls sv ix where
      mk z = let i = getIdx z in return (z :. activeBits i )
      step (z:.bs)
        | null bs   = return $ M.Done
        | otherwise = let (!h:ts) = bs in return $ M.Yield (ElmUnsetBit h (getIdx z `clearBit` h) z) (z:.ts)
      {-# INLINE [1] mk #-}
      {-# INLINE [1] step #-}
    {-# INLINE mkStream #-}
-}

{-
-- | Given a bitset, this terminal iterates through all possible
-- permutations of the active indices. If, say, indices @1,3,4@ are set,
-- then @1,3,4@, @1,4,3@, @3,1,4@, @3,4,1@, @4,1,3@, and @4,3,1@ are
-- produced. No bit is unset to the right of this terminal.
--
-- TODO the way the permutations are generated and provided to the user is
-- horribly slow

data SetPermutations = SetPermutations

instance (Element ls BitSet) => Element (ls :!: SetPermutations) BitSet where
    data Elm (ls :!: SetPermutations) BitSet = ElmSetPermutations ![Int] !BitSet !(Elm ls BitSet)
    type Arg (ls :!: SetPermutations)        = Arg ls :. [Int]
    getArg (ElmSetPermutations x _ ls)       = getArg ls :. x
    getIdx (ElmSetPermutations _ i _ )       = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance (Monad m, Element ls BitSet, MkStream m ls BitSet) => MkStream m (ls :!: SetPermutations) BitSet where
    mkStream (ls :!: SetPermutations) sv ix = M.flatten mk step Unknown $ mkStream ls sv ix where
      mk z = let i = getIdx z in return (z :. permutations (activeBits i))
      step (z:.ps)
        | null ps   = return $ M.Done
        | otherwise = let (!h:ts) = ps in return $ M.Yield (ElmSetPermutations h (getIdx z) z) (z:.ts)
      {-# INLINE [1] mk #-}
      {-# INLINE [1] step #-}
    {-# INLINE mkStream #-}
-}



